import numpy as np

# 准确率
def accuracy_score(y_true, y_pred):
    score = (y_pred == y_true).astype(np.int).sum()
    return score / len(y_true)

# 混淆矩阵（1为正类）
def confusion_matrix(y_true, y_pred, label=1):
    n = len(y_true)
    tp, fn, fp, tn = 0, 0, 0, 0
    for i in range(n):
        if y_true[i] == label and y_pred[i] == label:
            tp += 1
        elif y_true[i] == label and y_pred[i] != label:
            fn += 1
        elif y_true[i] != label and y_pred[i] == label:
            fp += 1
        elif y_true[i] != label and y_pred[i] != label:
            tn += 1
    return tp, fn, fp, tn

# 精确率
def precision(y_true, y_pred):
    tp, fn, fp, tn = confusion_matrix(y_true, y_pred)
    return tp / (tp + fp)

# 召回率
def recall(y_true, y_pred):
    tp, fn, fp, tn = confusion_matrix(y_true, y_pred)
    return tp / (tp + fn)

# f1-score
def f1_score(y_true, y_pred):
    tp, fn, fp, tn = confusion_matrix(y_true, y_pred)
    return 2 * tp / (2 * tp + fp + fn)

# jaccard系数
def jaccard_score(y_true, y_pred):
    n = len(set(y_pred))
    if n == 2:
        tp, fn, fp, tn = confusion_matrix(y_true, y_pred)
        return tp / (tp + fn + fn)
    else:
        score = []
        for i in range(n):
            tp, fn, fp, tn = confusion_matrix(y_true, y_pred, i)
            score.append(tp / (tp + fn + fn))
        return sum(score) / n

# roc-auc
def auc(y_true, y_score):
    # 标签为1的正类样本列表，列表元素是样本序号
    pos_sample = [i for i in range(len(y_true)) if y_true[i] == 1]
	# 标签为0的负类样本列表，列表元素是样本序号
    neg_sample = [i for i in range(len(y_true)) if y_true[i] == 0]

	# 遍历所有正负样本对，统计正样本预测值大于负样本预测值的样本总个数S
    s = 0
    for i in pos_sample:
        for j in neg_sample:
            if y_score[i] > y_score[j]:
                s += 1
			# 概率相等按0.5计算
            elif y_score[i] == y_score[j]:
                s += 0.5
    m = len(pos_sample)		# 正样本数量
    n = len(neg_sample)		# 负样本数量
    auc = s / (m * n)
    return auc 

# 决定系数（用于回归模型评估）
def r2_score(y_true, y_pred):
    # 计算误差平方和SSE
    sse = sum((y_pred - y_true) ** 2)
    # 计算总体离差平方和SST
    sst = sum((y_true - y_true.mean()) ** 2)
    # R2系数
    return 1 - sse / sst

# 聚类算法外部评价指标用a、b、c、d
def statis_abcd(y_true, y_pred):
    a, b, c, d = 0, 0, 0, 0
    n = len(y_true)
    for i in range(n):
        for j in range(i+1, n):
			# 原先同类，聚类后也是同一簇的样本对的数量
            if y_true[i] == y_true[j] and y_pred[i] == y_pred[j]:
                a += 1
			# 原先同类，聚类后不同簇的样本对的数量
            elif y_true[i] == y_true[j] and y_pred[i] != y_pred[j]:
                b += 1
			# 原先不同类，聚类后是同一簇的样本对的数量
            elif y_true[i] != y_true[j] and y_pred[i] == y_pred[j]:
                c += 1
			# 原先不同类，聚类后也不同簇的样本对的数量
            elif y_true[i] != y_true[j] and y_pred[i] != y_pred[j]:
                d += 1
    return a, b, c, d

# Fowlkes-Mallows Scores--聚类外部指标
def fmi(y_true, y_pred):
    a, b, c, d = statis_abcd(y_true, y_pred)
    return a / ((a + b) * (a + c)) ** 0.5

# 兰德指数(Rand index)--聚类外部指标
def ri(y_true, y_pred):
    a, b, c, d = statis_abcd(y_true, y_pred)
    return (a + d) / (a + b + c + d)


# 函数计算聚类算法的内部评价指标轮廓系数、DB指数和Dunn指数
# 输入：数据集X和聚类结果（标签向量）y
def internal_score(x, y):
    # 获取数据集维度，行row表示样本数量，列col表示特征数量
    row, col = x.shape

    # 计算所有样本与同簇其它样本的平均距离a

    # 设a为初始值0、长度为样本数的1维向量
    a = np.zeros(row, dtype=np.float)
    for i in range(row):
        # 筛选出同簇的样本
        temp_x = x[y == y[i]]
        # 计算样本i与同簇所有样本的欧氏距离
        dis = ((temp_x - x[i]) ** 2).sum(axis=1) ** 0.5
        # 计算平均距离，要在距离和中剔除样本i本身
        a[i] = dis.sum() / (len(temp_x) - 1)

    # 计算样本与其它簇的最小平均距离b

    # 设b初始值为0，长度为样本数的1维向量
    b = np.zeros(row, dtype=np.float)      
    # 求出所有标签类型放在集合中
    labels_set = set(y)
    # 标签类型数量，也是聚类中心的数量
    k = len(labels_set)
    for i in range(row):
        # 用集合运算--差集求除样本i所属簇号之外的簇集合
        temp_set = labels_set - {y[i]}
        min_dis = np.inf
        # 分别遍历其它簇所有样本，计算与这些簇内样本的欧氏距离的均值
        for j in temp_set:
            temp_x = x[y == j]
            dis = ((temp_x - x[i]) ** 2).sum(axis=1) ** 0.5
            dis = dis.mean()
			# 寻找最小值
            if min_dis > dis:
                min_dis = dis
        b[i] = min_dis

    # 计算簇中心，计算簇内样本到簇中心的平均距离C

    # 设centers初始值为0，长度为簇中心数的1维向量
    centers = np.zeros((k, col), dtype=np.float)
    c = np.zeros(k, dtype=np.float)
    for i in labels_set:
        temp_x = x[y == i]
		# 簇中心为簇内所有元素特征值的均值，是1维向量
        centers[i, :] = temp_x.mean(axis=0)
		# 簇中心到簇内所有样本的欧式距离
        dis = ((temp_x - centers[i]) ** 2).sum(axis=1) ** 0.5
        c[i] = dis.mean()
    
    # 计算簇中心之间最小距离
	# 遍历簇中心，计算每个簇中心与其它簇中心欧氏距离，并从中选取最小值
    dc_min = np.inf
    for i in range(k):
        for j in range(i+1, k):
			# 计算簇中心间欧式距离
            dis = ((centers[i] - centers[j]) ** 2).sum() ** 0.5
			# 寻找最小值
            if dc_min > dis:
                dc_min = dis

    # 计算轮廓系数sc
    sc = np.zeros(len(y), dtype=np.float)
    for i in range(len(y)):
        sc[i] = (b[i] - a[i]) / max(a[i], b[i])

    # 计算DB指数（Davies-Bouldin Index，DBI）
    # 遍历所有簇中心，计算出最大的几个Rij，然后求均值
    dbi = 0
    for i in range(k):
        max_r = 0               # 第i个簇中最大的Rij，初始值为0
        for j in labels_set:
			# 保证是不同簇之间进行运算
            if i != j:
                # 计算Rij
                dis_c = sum((centers[i] - centers[j]) ** 2) ** 0.5
                r = (c[i] + c[j]) / dis_c
                # 找到最大的Rij
                if max_r < r:
                    max_r = r
		# 累加Rij
        dbi += max_r
	# 计算平均值
    dbi /= k

    # 计算Dunn指数（Dunn Index，DI）
    di = dc_min / max(c)

    return sc, dbi, di
